// 002 u_c_camera_controler
// 27 may 2005

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

// -- control the azimuth / elevation / scale with the mouse

(*$r+*)

unit u_c_camera_controler;
  interface
    uses Classes // tShiftState
        , Controls // tMouseButton
        , ExtCtrls // tPanel, tPaintBox
        , Graphics
        , u_c_basic_object
        , u_2d_integer_point
        , u_2d_point
        ;

    const k_before_true= True;
          k_before_false= False;
          // -- minimum horizontal distance btw azimut and elevation
          k_azimuth_fat= 3;
          k_unknown_angle= pi / 20;

    type t_procedure= Procedure(p_before: Boolean);

         c_camera_controler= Class(c_basic_object)
                                    m_c_paintbox_ref: tPaintBox;
                                    _m_c_canvas_ref: tCanvas;

                                    // -- optimization: avoid recalculations
                                    m_2d_integer_origin: t_2d_integer_point;
                                    m_2d_origin: t_2d_point;
                                    m_2d_origin_x, m_2d_origin_y: t_2d_point;

                                    // -- the "unit" circle (half the paintbox)
                                    m_unit_width, m_unit_height: Integer;

                                    // -- the mouse in the world xy plane
                                    m_2d_azimuth: t_2d_point;
                                    // -- the mouse in the vertical
                                    m_2d_elevation: t_2d_point;

                                    m_azimuth, m_elevation, m_scale_ratio: Double;
                                    // -- when arrives in Oz // Omouse
                                    m_last_known_elevation: Double;
                                    m_is_azimuth, m_is_elevation: Boolean;

                                    m_on_change_controler: t_procedure;

                                    constructor create_camera_controler(p_name: String;
                                        p_c_paintbox_ref: tPaintBox;
                                        p_controler_ratio: Double;
                                        p_on_change_controler: t_procedure);

                                    function f_azimuth_angle(p_x, p_y: Double): Double;
                                    function f_elevation_angle(p_x, p_y: Double): Double;

                                    Procedure draw_unit_axes(p_color: Integer);
                                    procedure draw_all(p_color: Integer);

                                    procedure update_controler(p_x, p_y: Integer);

                                    procedure handle_mouse_down(Sender: TObject;
                                        p_mouse_button: TMouseButton;
                                        p_shift_state: TShiftState; p_x, p_y: Integer);
                                    procedure handle_mouse_move(Sender: TObject;
                                        p_shift_state: TShiftState; p_x, p_y: Integer);

                                    procedure initialize_position;

                                    Destructor Destroy; Override;
                                  end; // c_camera_controler_controler

  implementation
    uses SysUtils, Math, u_c_display
        , u_double
        , u_trigonometry
        ;

    // -- c_camera_controler

    constructor c_camera_controler.create_camera_controler(p_name: String;
        p_c_paintbox_ref: tPaintBox;
        p_controler_ratio: Double;
        p_on_change_controler: t_procedure);

      procedure position_controler;
        var l_length: Double;
        begin
          with m_c_paintbox_ref do
          begin
            m_unit_width:= Round(Width* p_controler_ratio/ 2);
            m_unit_height:= Round(Height* p_controler_ratio/ 2);

            m_2d_origin:= f_2d_point(Width / 2, Height / 2);
            m_2d_integer_origin:= f_2d_integer_point(Width div 2, Height div 2);

            l_length:= Sqrt(Sqr(Width)+ Sqr(Height))* 0.7* 0.5;
            // display(FloatToStr(l_length));
            m_2d_origin_x:= f_2d_point(- l_length* cos(pi/4), l_length* cos(pi/4));
            m_2d_origin_y:= f_2d_point(  l_length* cos(pi/4), l_length* cos(pi/4));

          end; // with m_c_paintbox_ref
        end; // position_controler

      begin // create_camera_controler
        Inherited create_basic_object(p_name);

        m_c_paintbox_ref:= p_c_paintbox_ref;
        _m_c_canvas_ref:= m_c_paintbox_ref.Canvas;

        position_controler;
        draw_all(clAqua);

        with m_c_paintbox_ref do
        begin
          OnMouseDown:= handle_mouse_down;
          OnMouseMove:= handle_mouse_move;
        end; // with m_c_paintbox_ref

        m_on_change_controler:= p_on_change_controler;

        m_scale_ratio:= 1.0;
      end; // create_camera_controler

    procedure c_camera_controler.draw_unit_axes(p_color: Integer);
      var l_save_pen_color: Integer;
      begin
        // -- vertical
        with m_c_paintbox_ref, _m_c_canvas_ref do
        begin
          l_save_pen_color:= Pen.Color;

          Pen.Color:= p_color;

          Ellipse(m_unit_width, m_unit_height, 3* m_unit_width, 3* m_unit_height);

          MoveTo(m_2d_integer_origin.m_x, 5);
          LineTo(m_2d_integer_origin.m_x, m_2d_integer_origin.m_y);

          MoveTo(m_2d_integer_origin.m_x, m_2d_integer_origin.m_y);
          LineTo(m_2d_integer_origin.m_x+ Round(m_2d_origin_x.m_x),
              m_2d_integer_origin.m_y+ Round(m_2d_origin_x.m_y));

          MoveTo(m_2d_integer_origin.m_x, m_2d_integer_origin.m_y);
          LineTo(m_2d_integer_origin.m_x+ Round(m_2d_origin_y.m_x),
              m_2d_integer_origin.m_y+ Round(m_2d_origin_y.m_x));

          Pen.Color:= l_save_pen_color;
        end; // with m_c_paintbox_ref, _m_c_canvas_ref
      end; // draw_unit_axes

    procedure c_camera_controler.draw_all(p_color: Integer);

      procedure draw_plot(p_x, p_y: Integer);
        const k_plot= 2;
        begin
          with _m_c_canvas_ref do
            Rectangle(p_x- k_plot, p_y- k_plot, p_x+ k_plot, p_y+ k_plot);
        end; // draw_plot

      var l_save_pen_color: Integer;

      begin // draw_all
        with _m_c_canvas_ref do
        begin
          l_save_pen_color:= Pen.Color;
          Pen.Color:= p_color;

          draw_unit_axes(clAqua);

          if m_2d_azimuth.m_x<> 0
            then begin
                MoveTo(m_2d_integer_origin.m_x, m_2d_integer_origin.m_y);
                LineTo(Round(m_2d_azimuth.m_x), Round(m_2d_azimuth.m_y));
                draw_plot(Round(m_2d_azimuth.m_x), Round(m_2d_azimuth.m_y));
              end;

          if m_2d_elevation.m_x<> 0
            then begin
                MoveTo(Round(m_2d_azimuth.m_x), Round(m_2d_azimuth.m_y));
                LineTo(Round(m_2d_elevation.m_x), Round(m_2d_elevation.m_y));
                draw_plot(Round(m_2d_elevation.m_x), Round(m_2d_elevation.m_y));
              end;

          Pen.Color:= l_save_pen_color;
        end; // with PaintBox1.Canvas, g_c_camera_controler
      end; // draw_all

    function c_camera_controler.f_azimuth_angle(p_x, p_y: Double): Double;
      var l_2d_origin_mouse: t_2d_point;

      function f_ellipse_radius: Double;
          // -- not totally correct: computes H+ (W-H)(1+ x%), but around 45 degrees wrong
        var l_vertical_cosine, l_vertical_angle: Double;
            l_angle: Double;
        begin
          l_vertical_cosine:= f_2d_cosine(l_2d_origin_mouse, f_2d_point(0, 1));
          l_vertical_angle:= f_arc_cosine(l_vertical_cosine);

          // -- if in 90..180, convert back to 0..90
          l_angle:= Abs(l_vertical_angle);
          if l_angle> pi/2
            then l_angle:= pi- l_angle;
          Result:= m_unit_height+ (m_unit_width- m_unit_height)* (l_angle/ (pi/ 2) );
          // display('radius '+ f_display_double(Result, 3, 1));
        end; // f_ellipse_radius

      var // l_cos_mouse_x: Double;
          l_cos_mouse_x, l_sin_mouse_x: Double;

      begin
        m_2d_azimuth:= f_2d_point(p_x, p_y);

        l_2d_origin_mouse:= f_2d_point(p_x- m_2d_origin.m_x, p_y- m_2d_origin.m_y);

        l_cos_mouse_x:= f_2d_cosine(l_2d_origin_mouse, m_2d_origin_x);
        l_sin_mouse_x:= f_2d_cosine(l_2d_origin_mouse, m_2d_origin_y);

        // display('ok '+ f_display_double(l_cos_mouse_x, 2, 3));
        //        Result:= ArcCos(l_cos_mouse_x);
        Result:= f_arc_cosine(l_cos_mouse_x);
        if l_cos_mouse_x>= 0
          then begin
              if l_sin_mouse_x>=0
                then
                else Result:= 2* pi- Result
            end
          else begin
              if l_sin_mouse_x>=0
                then
                else Result:= 2* pi- Result
            end;

        // -- we compare the mouse radius to the "unit ellipse"
        with l_2d_origin_mouse do
        begin
          if m_x= 0
            then m_scale_ratio:= Abs(m_y / m_unit_height)
            else
              if m_y= 0
                then m_scale_ratio:= Abs(m_x / m_unit_width)
                else begin
                    // m_scale_ratio:= Abs(m_y / ( m_unit_height  * (1+ (1/ m_unit_width)* m_y/ m_x)));
                    m_scale_ratio:= f_2d_norm(l_2d_origin_mouse)/ f_ellipse_radius;
                  end;
        end;
        // m_scale_ratio:= 1;
        // display(f_display_double(f_degree(Result), 4, 2)+ ' '+ f_display_double(m_scale_ratio, 2, 2));
      end; // f_azimuth_angle

    function c_camera_controler.f_elevation_angle(p_x, p_y: Double): Double;
      var l_unit_z, l_origin_z, l_2d_origin_mouse, l_origin_azimuth: t_2d_point;
          l_cos_z_azimuth, l_cos_mouse_azimuth: Double;
          l_angle_z_azimuth: Double;
      begin
        // -- make sure that the vertical control is always visible
        if p_y>= m_2d_azimuth.m_y- k_azimuth_fat
          then begin
              display('corr');
              p_y:= m_2d_azimuth.m_y- k_azimuth_fat;
            end;

        m_2d_elevation:= f_2d_point(p_x, p_y);

        l_unit_z:= f_2d_point(m_2d_origin.m_x, m_2d_origin.m_y- 1);

        if not f_2d_on_same_side(m_2d_origin, m_2d_azimuth, l_unit_z, f_2d_point(p_x, p_y))
          then begin
              display_bug_halt('below 0');
              Result:= 0;
            end
          else begin
              l_origin_z:= f_2d_point(0, -1);
              l_2d_origin_mouse:= f_2d_point(p_x- m_2d_origin.m_x, p_y- m_2d_origin.m_y);
              l_origin_azimuth:= f_2d_point(m_2d_azimuth.m_x- m_2d_origin.m_x, m_2d_azimuth.m_y- m_2d_origin.m_y);

              l_cos_z_azimuth:= f_2d_cosine(l_origin_z, l_origin_azimuth);
              l_cos_mouse_azimuth:= f_2d_cosine(l_2d_origin_mouse, l_origin_azimuth);

              l_angle_z_azimuth:= f_arc_cosine(l_cos_z_azimuth);
              Result:= f_arc_cosine(l_cos_mouse_azimuth);

              if Result> l_angle_z_azimuth
                then Result:= pi/ 2
                else begin
                    if (Abs(l_angle_z_azimuth)< k_unknown_angle)
                      then begin
                          Result:= m_last_known_elevation;
                          display('***');
                        end
                      else Result:= (pi / 2)* Result / l_angle_z_azimuth;
                  end;
            end;

         m_last_known_elevation:= Result;
         // display(f_display_double(f_degree(Result), 4, 2));
      end; // f_elevation_angle

    procedure c_camera_controler.update_controler(p_x, p_y: Integer);

      procedure set_azimuth(p_mouse_x, p_mouse_y: Integer);

        procedure correct_elevation;
          var l_2d_origin_azimuth: t_2d_point;
          begin
            // -- move the elevation point, keeping the horizontal elevation identical
            m_2d_elevation.m_x:= p_mouse_x;

            l_2d_origin_azimuth:= f_2d_point(m_2d_azimuth.m_x- m_2d_origin.m_x,
                m_2d_azimuth.m_y- m_2d_origin.m_y);
            m_2d_elevation.m_y:= m_2d_azimuth.m_y- f_2d_norm(l_2d_origin_azimuth)* f_tangent(m_elevation);
          end; // correct_elevation

        begin // set_azimuth
          // -- delete
          draw_all(clWhite);

          m_azimuth:= f_azimuth_angle(p_mouse_x, p_mouse_y);

          if m_2d_elevation.m_x<> 0
            then correct_elevation;

          // -- draw
          draw_all(clBlack);
        end; // set_azimuth

      procedure set_elevation(p_mouse_x, p_mouse_y: Integer);
        begin
          // -- delete
          draw_all(clWhite);

          // -- keep the elevation on a vertical line above the azimuth
          if m_2d_azimuth.m_x<> 0
            then p_mouse_x:= Round(m_2d_azimuth.m_x);

          m_elevation:= f_elevation_angle(p_mouse_x, p_mouse_y);

          // -- draw
          draw_all(clBlack);
        end; // set_elevation

      begin // update_controler
        if Assigned(m_on_change_controler)
          then m_on_change_controler(k_before_true);

        if m_is_azimuth
          then set_azimuth(p_x, p_y)
          else
            if m_is_elevation
              then set_elevation(p_x, p_y)
              else ;

        if Assigned(m_on_change_controler)
          then m_on_change_controler(k_before_false);

        // -- redraw the controler
        draw_all(clAqua);
      end; // update_controler

    procedure c_camera_controler.handle_mouse_down(Sender: TObject;
        p_mouse_button: TMouseButton;
        p_shift_state: TShiftState; p_x, p_y: Integer);

      function f_is_azimuth(p_mouse_x, p_mouse_y: Integer): Boolean;
        begin
          Result:= Abs(p_mouse_x- m_2d_azimuth.m_x)+ Abs(p_mouse_y- m_2d_azimuth.m_y)< 4;
          m_is_azimuth:= Result;
        end; // f_is_azimuth

      function f_is_elevation(p_mouse_x, p_mouse_y: Integer): Boolean;
        begin
          Result:= Abs(p_mouse_x- m_2d_elevation.m_x)+ Abs(p_mouse_y- m_2d_elevation.m_y)< 4;
          m_is_elevation:= Result;
        end; // f_is_elevation

      begin // handle_mouse_down
        if m_2d_azimuth.m_x= 0
          then begin
              // -- bogus clicks for the first construction
              m_is_azimuth:= true;
              update_controler(p_x, p_y);

              m_is_azimuth:= False;
              m_is_elevation:= True;
              update_controler(p_x, p_y- 5);
            end else

        // -- SIDE EFFECTS, catch the first of those
        if f_is_azimuth(p_x, p_y) or f_is_elevation(p_x, p_y)
          then begin
            end
      end; // handle_mouse_down

    procedure c_camera_controler.handle_mouse_move(Sender: TObject;
        p_shift_state: TShiftState; p_x, p_y: Integer);
      begin
        if ssLeft in p_shift_state
          then update_controler(p_x, p_y);
      end; // handle_mouse_move

    procedure c_camera_controler.initialize_position;
      var l_x, l_y: Integer;
      begin
        // -- bogus clicks for the first construction
        with m_c_paintbox_ref do
        begin
          l_x:= (Width* 7) div 13;
          l_y:= (Height* 5) div 6;
          m_is_azimuth:= True;
          update_controler(l_x, l_y);

          m_is_azimuth:= False;
          m_is_elevation:= True;
          update_controler(l_x, l_y- 20);
        end; // with m_c_paintbox_ref
      end; // initialize_position

    Destructor c_camera_controler.Destroy;
      begin
        Inherited;
      end; // Destroy

    end.


