// 002 u_c_perspective
// 27 may 2005

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_c_perspective;
  interface
    uses Graphics, u_c_basic_object
        , u_3d_point
        , u_2d_integer_point
        , u_2d_point
        ;

    type c_perspective= Class(c_basic_object)
                          m_c_canvas_ref: tCanvas;

                          m_3d_camera: t_3d_point;

                          s_11, s_12, s_13, s_21, s_22, s_23, s_32, s_33, s_43: Double;
                          _m_screen_width, _m_screen_height: integer;
                          _m_screen_center_x, _m_screen_center_y: integer;

                          // -- for the axis display
                          m_3d_origin: t_3d_point;
                          m_2d_origin: t_2d_integer_point;

                          // -- options
                          m_draw_color: Boolean;
                          m_draw_normal: Boolean;

                          constructor create_perspective(p_name: String; p_c_canvas: tCanvas);

                          procedure initialize_screen_position(p_center_x, p_center_y,
                              p_width, p_height: integer);
                          procedure initialise_camera_position(p_camera_x, p_camera_y, p_camera_z: Double;
                              p_scale: Double);

                          function f_project_3d_point(const pk_3d_point: t_3d_point): t_2d_point;
                          function f_project_3d_point_integer(const pk_3d_point: t_3d_point): t_2d_integer_point;

                          procedure draw_3d_segment(pk_3d_point_1, pk_3d_point_2: t_3d_point;
                              p_color: Integer);
                          procedure draw_2d_point(const pk_2d_point: t_2d_point; p_pen_color: Integer);
                          procedure draw_3d_point(const pk_3d_point: t_3d_point; p_pen_color: Integer);
                          Procedure draw_unit_axes(p_color: Integer);

                          Destructor Destroy; Override;
                        end; // c_perspective

    function f_round_2d_point(const pk_2d_point: t_2d_point): t_2d_integer_point;

  implementation
    uses SysUtils, Math, u_c_display
        , u_double
        ;

    function f_round_2d_point(const pk_2d_point: t_2d_point): t_2d_integer_point;
      begin
        Result.m_x:= Round(pk_2d_point.m_x);
        Result.m_y:= Round(pk_2d_point.m_y);
      end; // f_round_2d_point

    // -- c_perspective

    constructor c_perspective.create_perspective(p_name: String; p_c_canvas: tCanvas);
      begin
        Inherited create_basic_object(p_name);

        m_c_canvas_ref:= p_c_canvas;
        initialise_camera_position(10, 10, 10,   10);
      end; // create_perspective

    procedure c_perspective.initialize_screen_position(p_center_x, p_center_y,
          p_width, p_height: integer);
      begin
        _m_screen_width:= p_width;
        _m_screen_height:= p_height;

        _m_screen_center_x:= p_center_x;
        _m_screen_center_y:= p_center_y;
      end; // initialise_center_et_dimensions_ecran

    procedure c_perspective.initialise_camera_position(p_camera_x, p_camera_y, p_camera_z: Double;
          p_scale: Double);
        // -- compute the transformation coefficients
      var l_xy_radius, l_radius, l_d23: Double;
      begin
        l_xy_radius:= Sqrt(Sqr(p_camera_x)+ Sqr(p_camera_y));
        l_radius:= Sqrt(Sqr(p_camera_x)+ Sqr(p_camera_y)+ Sqr(p_camera_z));
        l_d23:= l_xy_radius* l_radius;

        s_11:= - p_camera_y* p_scale/ l_xy_radius;
        s_12:= - p_camera_x* p_camera_z* p_scale/ l_d23;
        s_13:= - p_camera_x/ l_radius;

        s_21:=   p_camera_x* p_scale/ l_xy_radius;
        s_22:= - p_camera_y* p_camera_z* p_scale/ l_d23;
        s_23:= - p_camera_y/ l_radius;

        s_32:= l_xy_radius* p_scale/ l_radius;
        s_33:= - p_camera_z/ l_radius;
        s_43:= l_radius;

        // -- no use ?
        with m_3d_camera do
        begin
          m_x:= p_camera_x;
          m_y:= p_camera_y;
          m_z:= p_camera_z;
        end; // with m_3d_camera

        normalize_3d_point(m_3d_camera);
      end; // initialise_point_de_vue

    function c_perspective.f_project_3d_point(const pk_3d_point: t_3d_point): t_2d_point;
      Var l_x_t, l_y_t, l_z_t: Double;
      begin
        l_z_t:= pk_3d_point.m_x* s_13+ pk_3d_point.m_y* s_23+ pk_3d_point.m_z* s_33+ s_43;
        if f_double_is_null(l_z_t)
          then begin
              display_bug_halt('z_nil');
              Exit;
            end;
        l_x_t:= (pk_3d_point.m_x* s_11+ pk_3d_point.m_y* s_21)/ l_z_t;
        l_y_t:= (pk_3d_point.m_x* s_12+ pk_3d_point.m_y* s_22+ pk_3d_point.m_z* s_32)/ l_z_t;

        // -- simple translation / scaling for screen
        Result.m_x:= 1.0* _m_screen_center_x+ l_x_t* _m_screen_width;
        Result.m_y:= 1.0* _m_screen_center_y- l_y_t* _m_screen_height;
      end; // f_project_3d_point

    function c_perspective.f_project_3d_point_integer(const pk_3d_point: t_3d_point): t_2d_integer_point;
      begin
        Result:= f_round_2d_point(f_project_3d_point(pk_3d_point));
      end; // f_project_3d_point_integer

    procedure c_perspective.draw_2d_point(const pk_2d_point: t_2d_point; p_pen_color: Integer);
      var l_save_pen_color: Integer;
      Begin
        with m_c_canvas_ref do
        begin
          l_save_pen_color:= Pen.Color;
          Pen.Color:= p_pen_color;

          with f_round_2d_point(pk_2d_point) do
            Ellipse(m_x- 2, m_y- 2, m_x+ 2, m_y+ 2);

          Pen.Color:= l_save_pen_color;
        end; // with m_c_canvas_ref
      end; // draw_2d_point

    procedure c_perspective.draw_3d_point(const pk_3d_point: t_3d_point; p_pen_color: Integer);
      begin
        draw_2d_point(f_project_3d_point(pk_3d_point), p_pen_color);
      end; // draw_3d_point

    procedure c_perspective.draw_3d_segment(pk_3d_point_1, pk_3d_point_2: t_3d_point;
        p_color: Integer);
      var l_save_pen_color: Integer;
      Begin
        with m_c_canvas_ref do
        begin
          l_save_pen_color:= Pen.Color;
          Pen.Color:= p_color;
          with f_project_3d_point_integer(pk_3d_point_1) do
            MoveTo(m_x, m_y);
          with f_project_3d_point_integer(pk_3d_point_2) do
            LineTo(m_x, m_y);
          Pen.Color:= l_save_pen_color;
        end; // with m_c_canvas_ref
      end; // draw_3d_segment

    Procedure c_perspective.draw_unit_axes(p_color: Integer);
      Begin
        m_3d_origin:= f_3d_point(0, 0, 0);
        m_2d_origin:= f_project_3d_point_integer(m_3d_origin);

        draw_3d_segment(m_3d_origin, f_3d_point(1, 0, 0), p_color);
        draw_3d_segment(m_3d_origin, f_3d_point(0, 1, 0), p_color);
        draw_3d_segment(m_3d_origin, f_3d_point(0, 0, 1), p_color);
      End; // draw_unit_axes

    Destructor c_perspective.Destroy;
      begin
        Inherited;
      end; // Destroy

    end.


