// 002 u_c_screen_controler
// 26 may 2005

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_c_screen_controler;
  interface
    uses Graphics
        , Classes // tShiftState
        , Controls // tMouseButton
        , StdCtrls
        , ComCtrls
        , ExtCtrls // tPanel, tPaintBox
        , u_c_basic_object;

    const k_plot= 3;
    type t_notify_change= Procedure;
         c_screen_controler= class(c_basic_object)
                               m_c_parent_panel_ref: tPanel;
                               m_c_paintbox: tPaintBox;
                               m_base_ratio: Double;

                               m_scale_start_x, m_scale_start_y: Integer;
                               m_scale_end_x, m_scale_end_y: Integer;

                               m_ratio_height, m_ratio_width: Double;
                               m_ratio_x, m_ratio_y: Double;

                               // -- the fixed figure representing the screen
                               m_base_x, m_base_y: Integer;
                               m_base_width, m_base_height: Integer;

                               // -- for moving (Right click)
                               m_mouse_x, m_mouse_y: Integer;

                               m_on_notify_change: t_notify_change;

                               Constructor create_screen_controler(p_name: String;
                                   p_c_parent_panel_ref: tPanel;
                                   p_base_ratio: Double;
                                   p_ratio_width, p_ratio_height: Double;
                                   p_ratio_x, p_ratio_y: Double;
                                   p_handle_scaler_change: t_notify_change);

                               function f_display_scale: String;
                               function f_display_ratios: String;
                               procedure draw_rectangle(p_scale_end_x, p_scale_end_y: Integer);
                               procedure handle_paint(Sender: TObject);

                               procedure handle_mouse_down(Sender: TObject;
                                   p_mouse_button: TMouseButton;
                                   p_shift_state: TShiftState; p_x, p_y: Integer);
                               procedure handle_mouse_move(Sender: TObject;
                                   p_shift_state: TShiftState; p_x, p_y: Integer);

                               Destructor Destroy; Override;
                             end; // c_screen_controler

  implementation
    Uses SysUtils, u_c_display;

    // -- c_screen_controler

    Constructor c_screen_controler.create_screen_controler(p_name: String;
        p_c_parent_panel_ref: tPanel;
        p_base_ratio: Double;
        p_ratio_width, p_ratio_height: Double;
        p_ratio_x, p_ratio_y: Double;
        p_handle_scaler_change: t_notify_change);

      procedure compute_figure;
        var l_figure_height, l_figure_width: Integer;
       begin
          with m_c_paintbox do
          begin
            m_base_width:= Round(Width* m_base_ratio);
            m_base_height:= Round(Height* m_base_ratio);
            m_base_x:= (Width- m_base_width) div 2;
            m_base_y:= (Height- m_base_height) div 2;

            l_figure_width:= Round(m_base_width* m_ratio_width);
            l_figure_height:= Round(m_base_height* m_ratio_height);

            m_scale_start_x:= Round(m_base_x+ m_base_width* m_ratio_x);
            m_scale_start_y:= Round(m_base_y+ m_base_height* m_ratio_y);
            m_scale_end_x:= m_scale_start_x+ l_figure_width;
            m_scale_end_y:= m_scale_start_y+ l_figure_height;
          end; // with m_c_paintbox
        end; // compute_figure

      begin // create_screen_controler
        Inherited create_basic_object(p_name);

        m_c_parent_panel_ref:= p_c_parent_panel_ref;
        m_c_parent_panel_ref.Color:= clWhite;
        m_c_parent_panel_ref.Caption:= '';

        m_base_ratio:= p_base_ratio;

        m_ratio_height:= p_ratio_height;
        m_ratio_width:= p_ratio_width;
        m_ratio_x:= p_ratio_x;
        m_ratio_y:= p_ratio_y;

        m_c_paintbox:= tPaintBox.Create(m_c_parent_panel_ref);
        with m_c_paintbox do
        begin
          Parent:= m_c_parent_panel_ref;
          Align:= alClient;
          OnPaint:= handle_paint;
          OnMouseDown:= handle_mouse_down;
          OnMouseMove:= handle_mouse_move;
        end; // with m_c_paintbox

        compute_figure;

        m_on_notify_change:= p_handle_scaler_change;
      end; // create_screen_controler

    function c_screen_controler.f_display_scale: String;
      begin
        Result:= Format('%3d %3d  %3d %3d', [m_scale_start_x, m_scale_start_y, m_scale_end_x, m_scale_end_y]);
      end; // f_display_scale

    function c_screen_controler.f_display_ratios: String;
      begin
        Result:= Format('%5g %5g   %5g %5g',
            [m_ratio_height, m_ratio_width, m_ratio_x, m_ratio_y]);
      end; // f_display_ratios

    procedure c_screen_controler.draw_rectangle(p_scale_end_x, p_scale_end_y: Integer);

      procedure display_plots;

        procedure display_plot(p_x, p_y: Integer);
          begin
            with m_c_paintbox.Canvas do
              Rectangle(p_x- k_plot, p_y- k_plot, p_x+ k_plot, p_y+ k_plot);
          end; // display_plot

        begin // display_plots
          display_plot(m_scale_start_x, m_scale_start_y);
          display_plot(m_scale_end_x, m_scale_start_y);
          display_plot(m_scale_start_x, m_scale_end_y);
          display_plot(m_scale_end_x, m_scale_end_y);
        end; // display_plots

      var l_save_brush_style: tBrushStyle;
          l_save_pen_style: tPenStyle;
          l_save_pen_mode: tPenMode;

      begin // draw_rectangle
        with m_c_paintbox.Canvas do
        begin
          l_save_brush_style:= Brush.Style;
          l_save_pen_style:= Pen.Style;
          l_save_pen_mode:= Pen.Mode;

          Brush.Style:= bsClear;
          // -- cannot color if dot
          // Pen.Color:= clRed;
          Pen.Style:= psDot;
          Pen.Mode:= pmNot;

          Rectangle(m_scale_start_x, m_scale_start_y, p_scale_end_x, p_scale_end_y);

          Pen.Style:= psSolid;
          Brush.Style:= bsSolid;
          display_plots;

          Brush.Style:= l_save_brush_style;
          Pen.Style:= l_save_pen_style;
          Pen.Mode:= l_save_pen_mode;

          m_scale_end_x:= p_scale_end_x; m_scale_end_y:= p_scale_end_y;

          // -- recompute the ratios for user
          m_ratio_width:= (m_scale_end_x- m_scale_start_x)/ m_base_width;
          m_ratio_height:= (m_scale_end_y- m_scale_start_y)/ m_base_height;

          m_ratio_x:= (m_scale_start_x- m_base_x)/ m_base_width;
          m_ratio_y:= (m_scale_start_y- m_base_y)/ m_base_height;
          // display(Format('%3g %3g  %3g %3g', [m_ratio_width, m_ratio_height, m_ratio_x, m_ratio_y]));
        end; // with m_c_paintbox.Canvas
      end; // draw_rectangle

    procedure c_screen_controler.handle_paint(Sender: TObject);
      begin
        with m_c_paintbox, Canvas do
        begin
          Rectangle(m_base_x, m_base_y, m_base_x+ m_base_width, m_base_y+ m_base_height);
          draw_rectangle(m_scale_end_x, m_scale_end_y);
        end; // with m_c_paintbox, Canvas
      end; // handle_paint

    procedure c_screen_controler.handle_mouse_down(Sender: TObject;
        p_mouse_button: TMouseButton;
        p_shift_state: TShiftState; p_x, p_y: Integer);
      begin
        m_mouse_x:= p_x; m_mouse_y:= p_y;
      end; // handle_mouse_down

    procedure c_screen_controler.handle_mouse_move(Sender: TObject;
        p_shift_state: TShiftState; p_x, p_y: Integer);

      function f_point_in_rectangle(p_x_1, p_y_1, p_x_2, p_y_2: Integer): Boolean;
        begin
          Result:= (p_x>= p_x_1) and (p_x<= p_x_2)
            and (p_y>= p_y_1) and (p_y<= p_y_2);
        end; // f_point_in_rectangle

      function f_is_north_west: Boolean;
        begin
          Result:= f_point_in_rectangle(m_scale_start_x- k_plot, m_scale_start_y- k_plot,
              m_scale_start_x+ k_plot, m_scale_start_y+ k_plot);
        end; // f_is_north_west

      function f_is_north_east: Boolean;
        begin
          Result:= f_point_in_rectangle(m_scale_end_x- k_plot, m_scale_start_y- k_plot,
              m_scale_end_x+ k_plot, m_scale_start_y+ k_plot);
        end; // f_is_north_east

      function f_is_south_east: Boolean;
        begin
          Result:= f_point_in_rectangle(m_scale_end_x- k_plot, m_scale_end_y- k_plot,
              m_scale_end_x+ k_plot, m_scale_end_y+ k_plot);
        end; // f_is_south_east

      function f_is_south_west: Boolean;
        begin
          Result:= f_point_in_rectangle(m_scale_start_x- k_plot, m_scale_end_y- k_plot, 
              m_scale_start_x+ k_plot, m_scale_end_y+ k_plot);
        end; // f_is_south_west

      function f_in_scale: Boolean;
        begin
          Result:= f_point_in_rectangle(m_scale_start_x+ k_plot, m_scale_start_y+ k_plot,
              m_scale_end_x- k_plot, m_scale_end_y- k_plot);
        end; // f_in_scale

      var l_dx, l_dy: Integer;
          l_changed: Boolean;

      begin // handle_mouse_move
        if ssLeft in p_shift_state
          then begin
              draw_rectangle(m_scale_end_x, m_scale_end_y);
              l_dx:= p_x- m_mouse_x;
              l_dy:= p_y- m_mouse_y;
              l_changed:= False;

              // display(Format('%3d %3d  ', [p_x, p_y])+ f_display_scale);
              if f_is_north_west
                then begin
                    // -- change north east corner
                    Inc(m_scale_start_x, l_dx);
                    Inc(m_scale_start_y, l_dy);
                    l_changed:= True;
                  end else
              if f_is_north_east
                then begin
                    Inc(m_scale_end_x, l_dx);
                    Inc(m_scale_start_y, l_dy);
                    l_changed:= True;
                  end else
              if f_is_south_east
                then begin
                    Inc(m_scale_end_x, l_dx);
                    Inc(m_scale_end_y, l_dy);
                    l_changed:= True;
                  end else
              if f_is_south_west
                then begin
                    Inc(m_scale_start_x, l_dx);
                    Inc(m_scale_end_y, l_dy);
                    l_changed:= True;
                  end else
              if f_in_scale
                then begin
                    Inc(m_scale_start_x, l_dx);
                    Inc(m_scale_start_y, l_dy);
                    Inc(m_scale_end_x, l_dx);
                    Inc(m_scale_end_y, l_dy);
                    l_changed:= True;
                  end else
                  ;

              draw_rectangle(m_scale_end_x, m_scale_end_y);
              if l_changed and Assigned(m_on_notify_change)
                then m_on_notify_change();
            end;

        m_mouse_x:= p_x; m_mouse_y:= p_y;
      end; // handle_mouse_move

    Destructor c_screen_controler.Destroy;
      begin
        Inherited;
      end; // Destroy

    begin // u_c_screen_controler
    end. // u_c_screen_controler
