// 002 u_2d_point
// 26 may 2005

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_2d_point;
  interface
    uses Graphics, u_c_basic_object;

    type t_2d_point= record
                       m_x, m_y: Double;
                     end; // t_2d_integer_point
         t_pt_2d_point= ^ t_2d_point;

    function f_2d_point(p_x, p_y: Double): t_2d_point;
    function f_zero_2d_point: t_2d_point;
    function f_2d_cross_product(const pk_2d_point_1, pk_2d_point_2: t_2d_point): Double;
    function f_2d_norm(const pk_2d_point: t_2d_point): Double;
    function f_normalize_2d_point(const pk_2d_point: t_2d_point): t_2d_point;

    function f_2d_cosine(const pk_2d_point_1, pk_2d_point_2: t_2d_point): Double;
    function f_2d_swap_xy(const pk_2d_point: t_2d_point): t_2d_point;
    // function f_2d_sine(const pk_2d_point_1, pk_2d_point_2: t_2d_point): Double;
    function f_2d_on_same_side(const pk_2d_line_point_1, pk_2d_line_point_2, pk_2d_point_1, pk_2d_point_2: t_2d_point): Boolean;

    function f_add_2d_point(const pk_2d_point_1, pk_2d_point_2: t_2d_point): t_2d_point;
    function f_subtract_2d_point(const pk_2d_point_1, pk_2d_point_2: t_2d_point): t_2d_point;
    function f_multiply_2d_point_by_scalar(const pk_2d_point: t_2d_point; p_double: Double): t_2d_point;
    function f_2d_are_identical(const pk_2d_point_1, pk_2d_point_2: t_2d_point): boolean;
    function f_2d_are_aligned(const pk_2d_point_1, pk_2d_point_2, pk_2d_point_3: t_2d_point): boolean;

    function f_display_2d_point(const pk_2d_point: t_2d_point): String;
    function f_rotate_2d_point(p_2d_point: t_2d_point; p_angle: Double): t_2d_point;

    procedure draw_2d_point(p_2d_point: t_2d_point; p_c_canvas: tCanvas; p_pen_color, p_radius: Integer);

  implementation
    uses SysUtils, u_c_display, u_double
       , u_trigonometry;

    function f_2d_point(p_x, p_y: Double): t_2d_point;
      begin
        Result.m_x:= p_x;
        Result.m_y:= p_y;
      end; // f_2d_point

    function f_zero_2d_point: t_2d_point;
      begin
        Result:= f_2d_point(0, 0);
      end; // f_zero_2d_point

    function f_2d_cross_product(const pk_2d_point_1, pk_2d_point_2: t_2d_point): Double;
      begin
        with pk_2d_point_1 do
          Result:= m_x* pk_2d_point_2.m_x+ m_y* pk_2d_point_2.m_y
      end; // f_2d_cross_product

    function f_2d_norm(const pk_2d_point: t_2d_point): Double;
      begin
        with pk_2d_point do
          Result:= Sqrt(Sqr(m_x)+ Sqr(m_y));
      end; // f_2d_norm

    function f_normalize_2d_point(const pk_2d_point: t_2d_point): t_2d_point;
      var l_norm: Double;
      begin
        l_norm:= f_2d_norm(pk_2d_point);
        if f_double_is_null(l_norm)
          then begin
              // display('NULL ');
              Result:= f_zero_2d_point;
            end
          else begin
              Result.m_x:= pk_2d_point.m_x/ l_norm;
              Result.m_y:= pk_2d_point.m_y/ l_norm;
           end;
      end; // f_normalize_2d_point

    function f_2d_cosine(const pk_2d_point_1, pk_2d_point_2: t_2d_point): Double;
      var l_2d_normalized_point_1, l_2d_normalized_point_2: t_2d_point;
      begin
        l_2d_normalized_point_1:= f_normalize_2d_point(pk_2d_point_1);
        l_2d_normalized_point_2:= f_normalize_2d_point(pk_2d_point_2);
        Result:= f_2d_cross_product(l_2d_normalized_point_1, l_2d_normalized_point_2);
      end; // f_2d_cosine

    function f_2d_swap_xy(const pk_2d_point: t_2d_point): t_2d_point;
      begin
        Result.m_x:= pk_2d_point.m_y;
        Result.m_y:= pk_2d_point.m_x;
      end; // f_2d_swap_xy

    function f_2d_on_same_side(const pk_2d_line_point_1, pk_2d_line_point_2, pk_2d_point_1, pk_2d_point_2: t_2d_point): Boolean;
      var l_slope: Double;

      function f(const pk_2d_point: t_2d_point): Double;
        begin
          Result:= (pk_2d_point.m_y- pk_2d_line_point_1.m_y)- (pk_2d_point.m_x- pk_2d_line_point_1.m_x)* l_slope;
        end; // f

      var l_value_1, l_value_2: Double;

      begin // f_2d_on_same_side
        l_slope:= pk_2d_line_point_2.m_x- pk_2d_line_point_1.m_x;

        if f_double_is_null(l_slope)
          then begin
              // -- the line is vertical: compare the points delta y
              Result:= (pk_2d_point_1.m_x- pk_2d_line_point_2.m_x)* (pk_2d_point_2.m_x- pk_2d_line_point_2.m_x)>= 0
            end
          else begin
              l_slope:= (pk_2d_line_point_2.m_y- pk_2d_line_point_1.m_y) / l_slope;

              l_value_1:= f(pk_2d_point_1);
              l_value_2:= f(pk_2d_point_2);
              // display(f_display_double(l_value_1, 3, 2)+ ' '+ f_display_double(l_value_2, 3, 2));
              Result:= l_value_1* l_value_2>= 0;
            end;
      end; // f_2d_on_same_side

    function f_add_2d_point(const pk_2d_point_1, pk_2d_point_2: t_2d_point): t_2d_point;
      begin
        with pk_2d_point_1 do
        begin
          Result.m_x:= m_x+ pk_2d_point_2.m_x;
          Result.m_y:= m_y+ pk_2d_point_2.m_y;
        end; // with pk_2d_point_1 do
      end; // add_2d_point

    function f_subtract_2d_point(const pk_2d_point_1, pk_2d_point_2: t_2d_point): t_2d_point;
      begin
        with pk_2d_point_1 do
        begin
          Result.m_x:= m_x- pk_2d_point_2.m_x;
          Result.m_y:= m_y- pk_2d_point_2.m_y;
        end; // with pk_2d_point_1 do
      end; // f_subtract_2d_point

    function f_multiply_2d_point_by_scalar(const pk_2d_point: t_2d_point; p_double: Double): t_2d_point;
      begin
        with pk_2d_point do
        begin
          Result.m_x:= m_x* p_double;
          Result.m_y:= m_y* p_double;
        end; // with pk_2d_point_1 do
      end; // f_multiply_2d_point_by_scalar

    function f_2d_are_identical(const pk_2d_point_1, pk_2d_point_2: t_2d_point): boolean;
      begin
        with pk_2d_point_1 do
          Result:= f_double_is_null(Abs(m_x- pk_2d_point_2.m_x)
              + Abs(m_y- pk_2d_point_2.m_y));
      end; // f_2d_are_identical

    function f_2d_are_aligned(const pk_2d_point_1, pk_2d_point_2, pk_2d_point_3: t_2d_point): boolean;
      var l_12, l_13: t_2d_point;
          l_alpha: Double;
      begin
        l_12:= f_subtract_2d_point(pk_2d_point_1, pk_2d_point_2);
        l_13:= f_subtract_2d_point(pk_2d_point_2, pk_2d_point_3);

        Result:= false;

        if f_double_is_null(l_12.m_y)
          then begin
              Result:= f_double_is_null(l_13.m_y);
            end
          else begin
              l_alpha:= l_13.m_y/ l_12.m_y;
              Result:= f_double_is_null(l_13.m_x- l_alpha* l_12.m_x);
            end;
      end; // f_2d_are_aligned

    function f_display_2d_point(const pk_2d_point: t_2d_point): String;
      begin
        with pk_2d_point do
          Result:= f_display_double(m_x, 5, 2)+ ' '+ f_display_double(m_y, 5, 2)
      end; // f_display_2d_point

    function f_rotate_2d_point(p_2d_point: t_2d_point; p_angle: Double): t_2d_point;
      var l_sine, l_cosine: Double;
      begin
        l_sine:= f_sine(p_angle); l_cosine:= f_cosine(p_angle);
        with p_2d_point do
        begin
          Result.m_x:= m_x* l_cosine+ m_y* l_sine;
          Result.m_y:= - m_x* l_sine+ m_y* l_cosine;
        end; // with p_2d_point
      end; // f_rotate_2d_point

    procedure draw_2d_point(p_2d_point: t_2d_point; p_c_canvas: tCanvas; p_pen_color, p_radius: Integer);
      var l_save_pen_color: Integer;
      begin
        with p_c_canvas do
        begin
          l_save_pen_color:= Pen.Color;
          Pen.Color:= p_pen_color;

          with p_2d_point do
            Ellipse(Round(m_x)- p_radius, Round(m_y)- p_radius,
                  Round(m_x)+ p_radius, Round(m_y)+ p_radius);

          Pen.Color:= l_save_pen_color;
        end;
      end; // draw_2d_point

    begin // u_c_2d_integer_point
    end. // u_c_2d_integer_point
