// 002 u_3d_point
// 26 may 2005

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_3d_point;
  interface

    type t_3d_point= record
                       m_x, m_y, m_z: Double;
                     end; // t_3d_point

    function f_3d_point(p_x, p_y, p_z: Double): t_3d_point;
    Function f_3d_clone(const pk_3d_point: t_3d_point): t_3d_point;
    function f_zero_3d_point: t_3d_point;
    procedure reset_3d_point(var pv_3d_point: t_3d_point);
    Function f_3d_cross_product(const pk_3d_point_1, pk_3d_point_2: t_3d_point): Double;
    function f_3d_external_product(const pk_3d_point_1, pk_3d_point_2: t_3d_point): t_3d_point;
    Function f_3d_point_norm(const pk_3d_point: t_3d_point): Double;
    function f_normalize_3d_point(const pk_3d_point: t_3d_point): t_3d_point;
    Procedure normalize_3d_point(var pv_3d_point: t_3d_point);

    function f_add_3d_point(const pk_3d_point_1, pk_3d_point_2: t_3d_point): t_3d_point;
    function f_subtract_3d_point(const pk_3d_point_1, pk_3d_point_2: t_3d_point): t_3d_point;
    function f_multiply_3d_point_by_scalar(const pk_3d_point: t_3d_point; p_double: Double): t_3d_point;
    function f_3d_are_identical(const pk_3d_point_1, pk_3d_point_2: t_3d_point): boolean;
    function f_3d_are_aligned(const pk_3d_point_1, pk_3d_point_2, pk_3d_point_3: t_3d_point): boolean;

    function f_display_3d_point(const pk_3d_point: t_3d_point): String;

  implementation
    uses SysUtils, u_c_display
        , u_double
        ;

    // -- t_3d_point

    function f_3d_point(p_x, p_y, p_z: Double): t_3d_point;
      begin
        Result.m_x:= p_x;
        Result.m_y:= p_y;
        Result.m_z:= p_z;
      end; // f_3d_point

    Function f_3d_clone(const pk_3d_point: t_3d_point): t_3d_point;
      begin
        with pk_3d_point do
          Result:= f_3d_point(m_x, m_y, m_z);
      end; // f_3d_clone

    function f_zero_3d_point: t_3d_point;
      begin
        Result:= f_3d_point(0, 0, 0);
      end; // f_zero_3d_point

    procedure reset_3d_point(var pv_3d_point: t_3d_point);
      begin
        FillChar(pv_3d_point, SizeOf(t_3d_point), 0);
      end; // reset_3d_point

    Function f_3d_cross_product(const pk_3d_point_1, pk_3d_point_2: t_3d_point): Double;
      begin
        with pk_3d_point_1 do
          Result:= m_x* pk_3d_point_2.m_x+ m_y* pk_3d_point_2.m_y+ m_z* pk_3d_point_2.m_z
      end; // f_3d_cross_product

    function f_3d_external_product(const pk_3d_point_1, pk_3d_point_2: t_3d_point): t_3d_point;
      var l_x, l_y, l_z: Double;
      begin
        with pk_3d_point_1 do
        begin
          l_x:= m_y* pk_3d_point_2.m_z- m_z* pk_3d_point_2.m_y;
          l_y:= m_z* pk_3d_point_2.m_x- m_x* pk_3d_point_2.m_z;
          l_z:= m_x* pk_3d_point_2.m_y- m_y* pk_3d_point_2.m_x;
        end; // with pk_3d_point_1
        Result:= f_3d_point(l_x, l_y, l_z);
      end; // f_3d_external_product

    Function f_3d_point_norm(const pk_3d_point: t_3d_point): Double;
      var l_cross: Double;
      begin
        with pk_3d_point do
          Result:= Sqrt(Sqr(m_x)+ Sqr(m_y)+ Sqr(m_z));
      end; // f_3d_norm

    function f_normalize_3d_point(const pk_3d_point: t_3d_point): t_3d_point;
      var l_norm: Double;
      begin
        l_norm:= f_3d_point_norm(pk_3d_point);

        if f_double_is_null(l_norm)
          then Result:= f_zero_3d_point
          else
            with pk_3d_point do
            begin
              Result.m_x:= m_x/ l_norm;
              Result.m_y:= m_y/ l_norm;
              Result.m_z:= m_z/ l_norm;
            end;
      end; // f_normalize_3d_point

    Procedure normalize_3d_point(var pv_3d_point: t_3d_point);
      begin
        pv_3d_point:= f_normalize_3d_point(pv_3d_point);
      end; // normalize_3d

    function f_add_3d_point(const pk_3d_point_1, pk_3d_point_2: t_3d_point): t_3d_point;
      begin
        with pk_3d_point_1 do
        begin
          Result.m_x:= m_x+ pk_3d_point_2.m_x;
          Result.m_y:= m_y+ pk_3d_point_2.m_y;
          Result.m_z:= m_z+ pk_3d_point_2.m_z;
        end; // with pk_3d_point_1 do
      end; // add_3d_point

    function f_subtract_3d_point(const pk_3d_point_1, pk_3d_point_2: t_3d_point): t_3d_point;
      begin
        with pk_3d_point_1 do
        begin
          Result.m_x:= m_x- pk_3d_point_2.m_x;
          Result.m_y:= m_y- pk_3d_point_2.m_y;
          Result.m_z:= m_z- pk_3d_point_2.m_z;
        end; // with pk_3d_point_1 do
      end; // f_subtract_3d_point

    function f_multiply_3d_point_by_scalar(const pk_3d_point: t_3d_point; p_double: Double): t_3d_point;
      begin
        with pk_3d_point do
        begin
          Result.m_x:= m_x* p_double;
          Result.m_y:= m_y* p_double;
          Result.m_z:= m_z* p_double;
        end; // with pk_3d_point_1 do
      end; // f_multiply_3d_point_by_scalar

    function f_3d_are_identical(const pk_3d_point_1, pk_3d_point_2: t_3d_point): boolean;
      begin
        with pk_3d_point_1 do
          Result:= f_double_is_null(Abs(m_x- pk_3d_point_2.m_x)
              + Abs(m_y- pk_3d_point_2.m_y)
              + Abs(m_z- pk_3d_point_2.m_z));
      end; // f_3d_are_identical

    function f_3d_are_aligned(const pk_3d_point_1, pk_3d_point_2, pk_3d_point_3: t_3d_point): boolean;
      var l_12, l_13: t_3d_point;
          l_alpha: Double;
      begin
        l_12:= f_subtract_3d_point(pk_3d_point_1, pk_3d_point_2);
        l_13:= f_subtract_3d_point(pk_3d_point_2, pk_3d_point_3);

        Result:= false;

        if f_double_is_null(l_12.m_z)
          then begin
              if f_double_is_null(l_13.m_z)
                then begin
                    if f_double_is_null(l_12.m_y)
                      then begin
                          Result:= f_double_is_null(l_13.m_y);
                        end
                      else begin
                          l_alpha:= l_13.m_y/ l_12.m_y;
                          Result:= f_double_is_null(l_13.m_x- l_alpha* l_12.m_x);
                        end;
                  end;
            end
          else begin
              l_alpha:= l_13.m_z/ l_12.m_z;
              Result:= f_double_is_null(l_13.m_x- l_alpha* l_12.m_x) and f_double_is_null(
                  l_13.m_y- l_alpha* l_12.m_y)
            end;
      end; // f_3d_are_aligned

    function f_display_3d_point(const pk_3d_point: t_3d_point): String;
      begin
        with pk_3d_point do
          // Result:= Format('%3g %3g %3g', [m_x, m_y, m_z]);
          Result:= f_display_double(m_x, 5, 2)
             + ' '+ f_display_double(m_y, 5, 2)
             + ' '+ f_display_double(m_z, 5, 2);
      end; // f_display_3d_point

    end. // principal


