// 002 u_c_3d_vertex_list
// 26 may 2005

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_c_3d_vertex_list;
  interface
    uses Classes, u_c_basic_object
        , u_3d_point, u_2d_integer_point
        , u_c_3d_vertex
        , u_c_perspective
        ;

    type c_3d_vertex_list= // "3d_vertex" list
                           Class(c_basic_object)
                             _m_c_3d_vertex_list: tStringList;
                             m_is_closed: Boolean;

                             Constructor create_3d_vertex_list(p_name: String);

                             function f_3d_vertex_count: Integer;
                             function f_c_3d_vertex(p_3d_vertex_index: Integer): c_3d_vertex;
                             function f_index_of(p_3d_vertex_name: String): Integer;
                             function f_c_find_by_3d_vertex(p_3d_vertex_name: String): c_3d_vertex;
                             procedure add_3d_vertex(p_3d_vertex_name: String; p_c_3d_vertex: c_3d_vertex);
                             function f_c_add_3d_vertex(p_3d_vertex_name: String;
                                 p_3d_point: t_3d_point): c_3d_vertex;
                             procedure display_3d_vertex_list;
                             function f_vertex_list_names: String;

                             procedure draw_vertex_list(p_c_perspective: c_perspective; p_pen_color: Integer);
                             procedure translate_3d_vertex_list(p_3d_delta: t_3d_point);

                             Destructor Destroy; Override;
                           end; // c_3d_vertex_list

  implementation
    uses Graphics, SysUtils, u_c_display;

    // -- c_3d_vertex_list

    Constructor c_3d_vertex_list.create_3d_vertex_list(p_name: String);
      begin
        Inherited create_basic_object(p_name);

        _m_c_3d_vertex_list:= tStringList.Create;
        m_is_closed:= True;
      end; // create_3d_vertex_line

    function c_3d_vertex_list.f_3d_vertex_count: Integer;
      begin
        Result:= _m_c_3d_vertex_list.Count;
      end; // f_3d_vertex_count

    function c_3d_vertex_list.f_c_3d_vertex(p_3d_vertex_index: Integer): c_3d_vertex;
      begin
        // -- WRAP AROUND: for Count, return point 0
        if (p_3d_vertex_index= f_3d_vertex_count) and m_is_closed
          then Result:= c_3d_vertex(_m_c_3d_vertex_list.Objects[0])
          else Result:= c_3d_vertex(_m_c_3d_vertex_list.Objects[p_3d_vertex_index]);
      end; //  f_c_3d_vertex

    function c_3d_vertex_list.f_index_of(p_3d_vertex_name: String): Integer;
      begin
        Result:= _m_c_3d_vertex_list.IndexOf(p_3d_vertex_name);
      end; // f_index_of

    function c_3d_vertex_list.f_c_find_by_3d_vertex(p_3d_vertex_name: String): c_3d_vertex;
      var l_index_of: Integer;
      begin
        l_index_of:= f_index_of(p_3d_vertex_name);
        if l_index_of< 0
          then Result:= Nil
          else Result:= c_3d_vertex(_m_c_3d_vertex_list.Objects[l_index_of]);
      end; // f_c_find_by_name

    procedure c_3d_vertex_list.add_3d_vertex(p_3d_vertex_name: String; p_c_3d_vertex: c_3d_vertex);
      begin
        _m_c_3d_vertex_list.AddObject(p_3d_vertex_name, p_c_3d_vertex);
      end; // add_3d_vertex

    function c_3d_vertex_list.f_c_add_3d_vertex(p_3d_vertex_name: String;
        p_3d_point: t_3d_point): c_3d_vertex;
      begin
        Result:= c_3d_vertex.create_3d_vertex(p_3d_vertex_name, p_3d_point);
        _m_c_3d_vertex_list.AddObject(p_3d_vertex_name, Result);
      end; // f_c_add_3d_vertex

    procedure c_3d_vertex_list.display_3d_vertex_list;
      var l_3d_vertex_index: Integer;
      begin
        display(m_name+ ' '+ IntToStr(f_3d_vertex_count));

        for l_3d_vertex_index:= 0 to f_3d_vertex_count- 1 do
          display(f_c_3d_vertex(l_3d_vertex_index).f_display_3d_vertex);
      end; // display_3d_vertex_list

    function c_3d_vertex_list.f_vertex_list_names: String;
      var l_3d_vertex_index: Integer;
      begin
        Result:= '';
        for l_3d_vertex_index:= 0 to f_3d_vertex_count- 1 do
          with f_c_3d_vertex(l_3d_vertex_index) do
            Result:= Result+ m_name+ ' ';
      end; // f_vertex_list_names

    procedure c_3d_vertex_list.draw_vertex_list(p_c_perspective: c_perspective; p_pen_color: Integer);
      var l_3d_vertex_index: Integer;
      begin
        with p_c_perspective do
        begin
          for l_3d_vertex_index:= 0 to f_3d_vertex_count- 2 do
            draw_3d_segment(f_c_3d_vertex(l_3d_vertex_index).m_3d_point,
                f_c_3d_vertex(l_3d_vertex_index+ 1).m_3d_point, p_pen_color);

          if m_is_closed
            then draw_3d_segment(f_c_3d_vertex(f_3d_vertex_count- 1).m_3d_point,
                f_c_3d_vertex(0).m_3d_point, p_pen_color);
        end; // with p_c_perspective
      end; // draw_vertex_list

    procedure c_3d_vertex_list.translate_3d_vertex_list(p_3d_delta: t_3d_point);
      var l_3d_vertex_index: Integer;
      begin
        for l_3d_vertex_index:= 0 to f_3d_vertex_count- 1 do
          f_c_3d_vertex(l_3d_vertex_index).translate_3d_vertex(p_3d_delta);
      end; // translate_3d_vertex_list

    Destructor c_3d_vertex_list.Destroy;
      var l_3d_vertex_index: Integer;
      begin
        for l_3d_vertex_index:= 0 to f_3d_vertex_count- 1 do
          f_c_3d_vertex(l_3d_vertex_index).Free;
        _m_c_3d_vertex_list.Free;

        Inherited;
      end; // Destroy

    begin // u_c_3d_vertex_list
    end. // u_c_3d_vertex_list
