// 002 u_c_facet_list
// 27 may 2005

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_c_facet_list;
  interface
    uses Classes, u_c_basic_object
        , u_c_3d_vertex_list
        , u_c_perspective
        , u_3d_point
        , u_2d_point
        ;

    type c_facet= // one "facet"
                  Class(c_basic_object)
                    m_c_3d_vertex_list: c_3d_vertex_list;

                    m_normal_sign: Integer;
                    m_3d_normal: t_3d_point;

                    m_facet_color: Integer;

                    Constructor create_facet(p_name: String; p_normal_sign: Integer);
                    function f_display_facet: String;
                    function f_c_self: c_facet;
                    procedure compute_normal;
                    procedure draw_facet(p_c_perspective: c_perspective; p_pen_color: Integer);
                    function f_3d_normal_end: t_3d_point;

                    function f_is_2d_point_in_2d_facet(p_2d_point: t_2d_point;
                        p_c_perspective: c_perspective): boolean;

                    procedure translate(p_dx, p_dy, p_dz: Double);
                    procedure translate_facet(p_3d_delta: t_3d_point);

                    Destructor Destroy; Override;
                  end; // c_facet

         c_facet_list= // "facet" list
                       Class(c_basic_object)
                         m_c_facet_list: tStringList;

                         Constructor create_facet_list(p_name: String);

                         function f_facet_count: Integer;
                         function f_c_facet(p_facet_index: Integer): c_facet;
                         function f_index_of(p_facet_name: String): Integer;
                         function f_c_find_by_facet(p_facet_name: String): c_facet;
                         procedure add_facet(p_facet_name: String; p_c_facet: c_facet);
                         function f_c_add_facet(p_facet_name: String; p_normal_sign: Integer): c_facet;
                         procedure display_facet_list;

                         function f_c_self: c_facet_list;

                         procedure compute_facet_normals;
                         procedure draw(p_c_perspective: c_perspective; p_pen_color: Integer);

                         function f_is_selected(p_2d_mouse: t_2d_point;
                             p_c_perspective: c_perspective): Boolean;
                         procedure translate(p_dx, p_dy, p_dz: Double);
                         procedure translate_facet_list(p_3d_delta: t_3d_point);
                         procedure display_unique_facet_list;
                         procedure set_position(p_3d_position: t_3d_point);

                         Destructor Destroy; Override;
                       end; // c_facet_list

  implementation
    uses SysUtils, Graphics
        , u_types_constants, u_characters, u_c_display
        , u_double
        ;

    // -- c_facet

    Constructor c_facet.create_facet(p_name: String;
        p_normal_sign: Integer);
      begin
        Inherited create_basic_object(p_name);

        m_c_3d_vertex_list:= c_3d_vertex_list.create_3d_vertex_list('v_'+ p_name);
        m_normal_sign:= p_normal_sign;
      end; // create_facet

    function c_facet.f_display_facet: String;
      begin
        Result:= Format('%-10s ', [m_name]);
      end; // f_display_facet

    function c_facet.f_c_self: c_facet;
      begin
        Result:= Self;
      end; // f_c_self

    procedure c_facet.compute_normal;
      var l_3d_vertex_1, l_3d_vertex_2, l_3d_vertex_3: t_3d_point;
          l_edge_12, l_edge_23: t_3d_point;
      begin
        with m_c_3d_vertex_list do
        begin
          l_3d_vertex_1:= f_c_3d_vertex(0).m_3d_point;
          l_3d_vertex_2:= f_c_3d_vertex(1).m_3d_point;
          l_3d_vertex_3:= f_c_3d_vertex(2).m_3d_point;
        end; // with m_c_3d_vertex_list

        l_edge_12:= f_subtract_3d_point(l_3d_vertex_1, l_3d_vertex_2);
        l_edge_23:= f_subtract_3d_point(l_3d_vertex_2, l_3d_vertex_3);

        m_3d_normal:= f_3d_external_product(l_edge_12, l_edge_23);
        if m_normal_sign= -1
          then m_3d_normal:= f_multiply_3d_point_by_scalar(m_3d_normal, -1);
        normalize_3d_point(m_3d_normal);
      end; // compute_normal

    function c_facet.f_is_2d_point_in_2d_facet(p_2d_point: t_2d_point;
        p_c_perspective: c_perspective): boolean;
        // -- compute the intersection count with an horizontal line
      const k_Ox_infinite= 1000;
      var l_2d_Ox_infinite: t_2d_point;
          l_intersection_count: integer;
      function f_intersection_count(p_2d_one, p_2d_two: t_2d_point): integer;
        var l_2d_one_two, l_2d_one_point, l_2d_point_infinite: t_2d_point;
            l_u: Double;
            l_2d_intersection: t_2d_point;
            l_2d_intersection_point: t_2d_point;
        begin
          l_2d_one_two:= f_subtract_2d_point(p_2d_two, p_2d_one);
          l_2d_one_point:= f_subtract_2d_point(p_2d_point, p_2d_one);
          l_2d_point_infinite:= f_subtract_2d_point(l_2d_Ox_infinite, p_2d_point);

          l_u:= (l_2d_one_point.m_x* l_2d_point_infinite.m_y -l_2d_one_point.m_y* l_2d_point_infinite.m_x)
            / (l_2d_one_two.m_x* l_2d_point_infinite.m_y -l_2d_one_two.m_y* l_2d_point_infinite.m_x);

          l_2d_intersection.m_x:= p_2d_one.m_x+ l_u* l_2d_one_two.m_x;
          l_2d_intersection.m_y:= p_2d_one.m_y+ l_u* l_2d_one_two.m_y;

          // -- if the intersection is between the two border point
          // -- also check that this intersection is from the point to the infinite
          if (l_u>= 0) and (l_u<= 1)
            then begin
                l_2d_intersection_point:= f_subtract_2d_point(l_2d_intersection, p_2d_point);
                if f_2d_cross_product(l_2d_intersection_point, l_2d_point_infinite)> 0
                  then Result:= 1
                  else Result:= 0;
              end
            else Result:= 0;
        end; // f_intersection_count

      var l_vertex_index: Integer;

      begin // f_is_2d_point_in_2d_facet
        l_intersection_count:= 0;

        // -- compute the Ox infinite point
        l_2d_Ox_infinite.m_x:= k_Ox_infinite;
        l_2d_Ox_infinite.m_y:= 0;
        with m_c_3d_vertex_list, p_c_perspective do
        begin
          // display('3d_vx_count '+ IntToStr(f_3d_vertex_count));
          for l_vertex_index:= 0 to f_3d_vertex_count- 2 do
          begin
            Inc(l_intersection_count,
                f_intersection_count(f_project_3d_point(f_c_3d_vertex(l_vertex_index).m_3d_point),
                    f_project_3d_point(f_c_3d_vertex(l_vertex_index+ 1).m_3d_point)));
          end;
        end;
        Result:= odd(l_intersection_count);
      end; // f_is_2d_point_in_2d_facet

    procedure c_facet.draw_facet(p_c_perspective: c_perspective; p_pen_color: Integer);
      var l_cross_product: Double;
          l_is_visible: Boolean;
          l_pen_color: Integer;
          l_3d_middle: t_3d_point;
          l_2d_middle_projection: t_2d_point;
      begin
        l_cross_product:= f_3d_cross_product(m_3d_normal, p_c_perspective.m_3d_camera);

        // l_is_visible:= l_cross_product>= 0;
        l_is_visible:= l_cross_product>= -0.10;
(*
        if not l_is_visible
          then display('inv '+ m_c_3d_vertex_list.f_c_3d_vertex(0).m_name+ ' '+ f_display_double(l_cross_product, 3, 2))
          else display('vis '+ m_c_3d_vertex_list.f_c_3d_vertex(0).m_name+ ' '+ f_display_double(l_cross_product, 3, 2));
*)

        if l_is_visible
          then begin
              if (p_pen_color= clWhite)
                  or (m_c_3d_vertex_list.f_3d_vertex_count<> 4)
                  or not p_c_perspective.m_draw_color
                then m_c_3d_vertex_list.draw_vertex_list(p_c_perspective, p_pen_color)
                else begin
                    with m_c_3d_vertex_list do
                    begin
                      l_3d_middle:= f_add_3d_point(f_c_3d_vertex(0).m_3d_point, f_c_3d_vertex(2).m_3d_point);
                      l_3d_middle:= f_multiply_3d_point_by_scalar(l_3d_middle, 0.5);
                      l_2d_middle_projection:= p_c_perspective.f_project_3d_point(l_3d_middle);
                    end;

                    m_c_3d_vertex_list.draw_vertex_list(p_c_perspective, clOlive);
                    p_c_perspective.m_c_canvas_ref.Brush.Color:= clYellow;
                    with l_2d_middle_projection do
                      p_c_perspective.m_c_canvas_ref.FloodFill(Round(m_x), Round(m_y), clOlive, fsBorder);
                    p_c_perspective.m_c_canvas_ref.Brush.Color:= clWhite;

                    m_c_3d_vertex_list.draw_vertex_list(p_c_perspective, p_pen_color)
                  end;
            end;

        // -- debug
        if p_c_perspective.m_draw_normal
          then
            begin
              if p_pen_color= clBlack
                then l_pen_color:= clRed
                else l_pen_color:= p_pen_color;

              with p_c_perspective do
              begin
                draw_3d_segment(m_c_3d_vertex_list.f_c_3d_vertex(0).m_3d_point,
                  f_3d_normal_end, l_pen_color);
              end;
            end;
      end; // draw_facet

    function c_facet.f_3d_normal_end: t_3d_point;
        // -- for debug: visualize the normals
      var l_3d_vertex_1: t_3d_point;
      begin
        with m_c_3d_vertex_list do
          l_3d_vertex_1:= f_c_3d_vertex(0).m_3d_point;
        Result:= f_add_3d_point(f_multiply_3d_point_by_scalar(f_3d_clone(m_3d_normal), 0.1), l_3d_vertex_1);
      end; // f_3d_normal_end

    procedure c_facet.translate(p_dx, p_dy, p_dz: Double);
      begin
        m_c_3d_vertex_list.translate_3d_vertex_list(f_3d_point(p_dx, p_dy, p_dz))
      end; // translate

    procedure c_facet.translate_facet(p_3d_delta: t_3d_point);
      begin
        m_c_3d_vertex_list.translate_3d_vertex_list(p_3d_delta);
      end; // translate_facet

    Destructor c_facet.Destroy;
      begin
        m_c_3d_vertex_list.Free;

        InHerited;
      end; // Destroy

    // -- c_facet_list

    Constructor c_facet_list.create_facet_list(p_name: String);
      begin
        Inherited create_basic_object(p_name);

        m_c_facet_list:= tStringList.Create;
      end; // create_facet_line

    function c_facet_list.f_facet_count: Integer;
      begin
        Result:= m_c_facet_list.Count;
      end; // f_facet_count

    function c_facet_list.f_c_facet(p_facet_index: Integer): c_facet;
      begin
        Result:= c_facet(m_c_facet_list.Objects[p_facet_index]);
      end; //  f_c_facet

    function c_facet_list.f_index_of(p_facet_name: String): Integer;
      begin
        Result:= m_c_facet_list.IndexOf(p_facet_name);
      end; // f_index_of

    function c_facet_list.f_c_find_by_facet(p_facet_name: String): c_facet;
      var l_index_of: Integer;
      begin
        l_index_of:= f_index_of(p_facet_name);
        if l_index_of< 0
          then Result:= Nil
          else Result:= c_facet(m_c_facet_list.Objects[l_index_of]);
      end; // f_c_find_by_name

    procedure c_facet_list.add_facet(p_facet_name: String; p_c_facet: c_facet);
      begin
        m_c_facet_list.AddObject(p_facet_name, p_c_facet);
      end; // add_facet

    function c_facet_list.f_c_add_facet(p_facet_name: String; p_normal_sign: Integer): c_facet;
      begin
        Result:= c_facet.create_facet(p_facet_name, p_normal_sign);
        m_c_facet_list.AddObject(p_facet_name, Result);
      end; // f_c_add_facet

    procedure c_facet_list.display_facet_list;
      var l_facet_index: Integer;
      begin
        display(m_name+ ' '+ IntToStr(f_facet_count));

        for l_facet_index:= 0 to f_facet_count- 1 do
          display(f_c_facet(l_facet_index).f_display_facet);
      end; // display_facet_list

    function c_facet_list.f_c_self: c_facet_list;
      begin
        Result:= Self;
      end; // f_c_self

    procedure c_facet_list.compute_facet_normals;
      var l_facet_index: Integer;
      begin
        for l_facet_index:= 0 to f_facet_count- 1 do
          f_c_facet(l_facet_index).compute_normal;
      end; // compute_facet_normals

    procedure c_facet_list.draw(p_c_perspective: c_perspective; p_pen_color: Integer);
      var l_facet_index: Integer;
      begin
        for l_facet_index:= 0 to f_facet_count- 1 do
          f_c_facet(l_facet_index).draw_facet(p_c_perspective, p_pen_color);
      end; // draw

    function c_facet_list.f_is_selected(p_2d_mouse: t_2d_point;
        p_c_perspective: c_perspective): Boolean;
        // -- is the mouse in the polygon
        // -- only checks for "rectangles"
      var l_facet_index: Integer;
      begin
        Result:= False;
        for l_facet_index:= 0 to f_facet_count- 1 do
          with f_c_facet(l_facet_index) do
            if f_is_2d_point_in_2d_facet(p_2d_mouse, p_c_perspective)
              then begin
                  Result:= True;
                  Break;
                end;
      end; // f_is_selected

    procedure c_facet_list.translate(p_dx, p_dy, p_dz: Double);
      var l_facet_index: Integer;
      begin
        for l_facet_index:= 0 to f_facet_count- 1 do
          f_c_facet(l_facet_index).translate(p_dx, p_dy, p_dz);
      end; // translate

    procedure c_facet_list.translate_facet_list(p_3d_delta: t_3d_point);
      var l_facet_index: Integer;
      begin
        for l_facet_index:= 0 to f_facet_count- 1 do
          f_c_facet(l_facet_index).translate_facet(p_3d_delta);
      end; // translate_facet_list

    procedure c_facet_list.display_unique_facet_list;
      var l_c_list: tStringList;
          l_facet_index: Integer;
          l_result: String;
          l_vertex_index: Integer;
          l_count_by_line: Integer;
      begin
        l_c_list:= tStringList.Create;
        l_count_by_line:= 0;

        for l_facet_index:= 0 to f_facet_count- 1 do
          with f_c_facet(l_facet_index).m_c_3d_vertex_list do
            for l_vertex_index:= 0 to f_3d_vertex_count- 1 do
              with f_c_3d_vertex(l_vertex_index) do
                if l_c_list.IndexOf(f_c_self.m_name)< 0
                  then begin
                      l_c_list.Add(f_c_self.m_name);
                      l_result:= l_result+ m_name+ ' '+ f_display_3d_point(f_c_self.m_3d_point)+ ', ';
                      Inc(l_count_by_line);
                      if (l_count_by_line mod 4)= 0
                        then l_result:= l_result+ k_new_line;
                    end;

        display(l_result);
        l_c_list.Free;
      end; // display_unique_facet_list

    procedure c_facet_list.set_position(p_3d_position: t_3d_point);
      var l_c_3d_delta: t_3d_point;
          l_3d_vertex_0: t_3d_point;
      begin
        // -- get the first
        l_3d_vertex_0:= f_c_facet(0).m_c_3d_vertex_list.f_c_3d_vertex(0).m_3d_point;
        l_c_3d_delta:= f_subtract_3d_point(p_3d_position, l_3d_vertex_0);

        translate_facet_list(l_c_3d_delta);
      end; // set_position

    Destructor c_facet_list.Destroy;
      var l_facet_index: Integer;
      begin
        for l_facet_index:= 0 to f_facet_count- 1 do
          f_c_facet(l_facet_index).Free;
        m_c_facet_list.Free;

        Inherited;
      end; // Destroy

    begin // u_c_facet_list
    end. // u_c_facet_list
