// 002 u_c_scene
// 28 may 2005

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_c_scene;
  interface
    uses Graphics
        , Classes // tShiftState
        , Controls // tMouseButton
        , StdCtrls
        , ComCtrls
        , ExtCtrls // tPanel, tPaintBox
        , u_c_basic_object
        , u_c_volume_list, u_c_perspective
        , u_c_facet_list
        ;

    const k_plot= 3;
    type t_direction= (e_unknown_direction, e_x_direction, e_y_direction, e_z_direction);

         t_notify_change= Procedure;
         c_scene= class(c_basic_object)
                    m_c_scene_paintbox_ref: tPaintBox;
                    m_c_volume_list_ref: c_volume_list;
                    m_c_perspective_ref: c_perspective;

                    m_direction: t_direction;
                    m_on_selection_change_event: t_notify_change;

                    m_c_selected_volume: c_facet_list;

                    Constructor create_scene(p_name: String;
                        p_c_scene_paintbox_ref: tPaintBox;
                        p_c_volume_list_ref: c_volume_list;
                        p_c_perspective_ref: c_perspective;
                        p_on_selection_change_event: t_notify_change);

                    procedure set_direction(p_direction: t_direction);

                    procedure handle_paint(Sender: TObject);
                    procedure handle_mouse_down(Sender: TObject;
                        p_mouse_button: TMouseButton;
                        p_shift_state: TShiftState; p_x, p_y: Integer);
                    procedure handle_mouse_move(Sender: TObject;
                        p_shift_state: TShiftState; p_x, p_y: Integer);

                    Destructor Destroy; Override;
                  end; // c_scene

    function f_direction_name(p_direction: t_direction): String;

  implementation
    Uses SysUtils, Math, TypInfo
        , u_c_display
        , u_trigonometry, u_double
        , u_2d_point
        , u_3d_point
        ;

    function f_direction_name(p_direction: t_direction): String;
      begin
        Result:= GetEnumName(TypeInfo(t_direction), Integer(p_direction));
        Delete(Result, 1, 2);
        Delete(Result, Length(Result)- Length('_direction')+ 1, Length('_direction'));
      end; // f_direction_name

    // -- c_scene

    Constructor c_scene.create_scene(p_name: String;
        p_c_scene_paintbox_ref: tPaintBox;
        p_c_volume_list_ref: c_volume_list;
        p_c_perspective_ref: c_perspective;
        p_on_selection_change_event: t_notify_change);
      begin
        Inherited create_basic_object(p_name);

        m_c_scene_paintbox_ref:= p_c_scene_paintbox_ref;
        with m_c_scene_paintbox_ref do
        begin
          OnPaint:= handle_paint;
          OnMouseDown:= handle_mouse_down;
          OnMouseMove:= handle_mouse_move;
        end; // with m_c_paintbox

        m_c_volume_list_ref:= p_c_volume_list_ref;
        m_c_perspective_ref:= p_c_perspective_ref;

        m_on_selection_change_event:= p_on_selection_change_event;
      end; // create_scene

    procedure c_scene.set_direction(p_direction: t_direction);
      begin
        m_direction:= p_direction;
      end; // set_direction

    procedure c_scene.handle_paint(Sender: TObject);
      begin
        with m_c_scene_paintbox_ref, Canvas do
        begin
(*
          Rectangle(m_base_x, m_base_y, m_base_x+ m_base_width, m_base_y+ m_base_height);
          draw_rectangle(m_scale_end_x, m_scale_end_y);
*)
        end; // with m_c_scene_paintbox_ref, Canvas
      end; // handle_paint

    var s_mouse_start_x, s_mouse_start_y: Integer;
        s_slow_dx, s_slow_dy: Integer;

    procedure c_scene.handle_mouse_down(Sender: TObject;
        p_mouse_button: TMouseButton;
        p_shift_state: TShiftState; p_x, p_y: Integer);
      begin
        display('> c_scene.mouse_down');

        // -- find which volume was clicked
        m_c_selected_volume:= m_c_volume_list_ref.f_c_find_selected_volume(f_2d_point(p_x, p_y),
            m_c_perspective_ref);
        s_mouse_start_x:= p_x; s_mouse_start_y:= p_y;
        s_slow_dx:= 0; s_slow_dy:= 0;

        If Assigned(m_on_selection_change_event)
          then m_on_selection_change_event();

        if m_c_selected_volume= Nil
          then display('< c_scene.mouse_down. No_sel ')
          else display('< c_scene.mouse_down. '+ m_c_selected_volume.m_name);
      end; // handle_mouse_down

    var s_mouse_count: Integer= 0;

    procedure c_scene.handle_mouse_move(Sender: TObject;
        p_shift_state: TShiftState; p_x, p_y: Integer);
      var l_dx, l_dy: Integer;
          l_2d_normalized_move: t_2d_point;
          l_2d_origin, l_2d_Ox, l_2d_Oy, l_2d_Oz: t_2d_point;
          l_cos_x, l_cos_y, l_cos_z: Double;

      procedure compute_cosines;
        begin
          l_2d_normalized_move:= f_normalize_2d_point(f_2d_point(l_dx, l_dy));

          l_2d_origin:= m_c_perspective_ref.f_project_3d_point(f_3d_point(0, 0, 0));

          l_2d_Ox:= m_c_perspective_ref.f_project_3d_point(f_3d_point(1, 0, 0));
          l_2d_Oy:= m_c_perspective_ref.f_project_3d_point(f_3d_point(0, 1, 0));
          l_2d_Oz:= m_c_perspective_ref.f_project_3d_point(f_3d_point(0, 0, 1));

          l_2d_Ox:= f_normalize_2d_point(f_subtract_2d_point(l_2d_ox, l_2d_origin));
          l_2d_Oy:= f_normalize_2d_point(f_subtract_2d_point(l_2d_oy, l_2d_origin));
          l_2d_Oz:= f_normalize_2d_point(f_subtract_2d_point(l_2d_oz, l_2d_origin));

          l_cos_x:= f_2d_cross_product(l_2d_normalized_move, l_2d_Ox);
          l_cos_y:= f_2d_cross_product(l_2d_normalized_move, l_2d_Oy);
          l_cos_z:= f_2d_cross_product(l_2d_normalized_move, l_2d_Oz);
        end; // compute_cosines

      procedure approximate_direction;
          // -- trial to move in Oxyz direction closest to the mouse move
        var l_abs_cos_x, l_abs_cos_y, l_abs_cos_z: Double;
            l_max: Double;
            l_is_x, l_is_y, l_is_z: Boolean;
        begin
          display('> approx');
          // -- accumulate to avoid (1, 0) or (0, 1) moves
          Inc(s_mouse_count);
          if s_mouse_count mod 4<> 0
            then Exit;

          l_dx:= p_x- s_mouse_start_x;
          l_dy:= p_y- s_mouse_start_y;

          if Abs(l_dx)+ Abs(l_dy)= 0
            then begin
                display('< approx. Exit');
                Exit;
              end
            else ; // display('move '+ IntToStr(l_dx)+ ' '+ IntToStr(l_dy));

          compute_cosines;

          l_abs_cos_x:= Abs(l_cos_x);
          l_abs_cos_y:= Abs(l_cos_y);
          l_abs_cos_z:= Abs(l_cos_z);
          l_max:= Max(l_abs_cos_x, Max(l_abs_cos_y, l_abs_cos_z));

          display('cos '+
              f_display_double(l_cos_x, 2, 2)
              + ' '+ f_display_double(l_cos_y, 2, 2)
              + ' '+ f_display_double(l_cos_z, 2, 2)
              + ' max '+ f_display_double(l_max, 2, 2));

          l_is_x:= False; l_is_y:= false; l_is_z:= false;
          if l_max= l_abs_cos_x
            then l_is_x:= True
            else
              if l_max= l_abs_cos_y
                then l_is_y:= True
                else
                  if l_max= l_abs_cos_z
                    then l_is_z:= True
                    else display_bug_halt('no_max');

          m_c_volume_list_ref.draw(m_c_perspective_ref, clWhite);

          if l_is_x
            then begin
                m_c_selected_volume.translate(l_dx* l_cos_x, 0, 0);
                display('X '+ f_display_double(l_dx* l_cos_x, 2, 2));
              end
            else
              if l_is_y
                then begin
                    m_c_selected_volume.translate(0, l_dx, 0);
                    display('Y');
                  end
                else begin
                    m_c_selected_volume.translate(0, 0, - l_dy);
                    display('Z');
                  end;

          m_c_volume_list_ref.draw(m_c_perspective_ref, clBlack);

          s_mouse_start_x:= p_x; s_mouse_start_y:= p_y;
          display('< approx');
        end; // approximate_direction

      procedure exact_direction;

        procedure move_direct;
          const k_slow= 3;
          var l_move_norm: Double;
          begin
            l_dx:= p_x- s_mouse_start_x;
            l_dy:= p_y- s_mouse_start_y;
            display(IntToStr(Abs(l_dx)+ Abs(l_dy)));

            if (Abs(l_dx)+ Abs(l_dy))> 2
              then begin
                  l_dx:= l_dx div k_slow;
                  l_dy:= l_dy div k_slow;

                  compute_cosines;
                  l_move_norm:= f_2d_norm(f_2d_point(l_dx, l_dy));

                  m_c_volume_list_ref.draw(m_c_perspective_ref, clWhite);

                  if m_direction= e_x_direction
                    then begin
                        m_c_selected_volume.translate(l_move_norm* l_cos_x/ 10, 0, 0);
                        display('dir_X '+ f_display_double(l_move_norm* l_cos_x, 2, 2));
                      end
                    else
                      if m_direction= e_y_direction
                        then begin
                            m_c_selected_volume.translate(0, l_move_norm* l_cos_y/ 10, 0);
                            display('dir_Y');
                          end
                        else begin
                            m_c_selected_volume.translate(0, 0, - l_dy/ 10);
                            display('dir_Z');
                          end;
                  s_mouse_start_x:= p_x; s_mouse_start_y:= p_y;

                  m_c_perspective_ref.draw_unit_axes(clBlack);
                  m_c_volume_list_ref.draw(m_c_perspective_ref, clBlack);
                end;
          end; // move_direct

        procedure cumulate_move;
            // -- accumulate the integer moves, but only change position if reaches a
            // -- threshold
          const k_slow= 10;
          var l_move_norm: Double;
          begin
            l_dx:= p_x- s_mouse_start_x;
            l_dy:= p_y- s_mouse_start_y;
            display(IntToStr(Abs(l_dx)+ Abs(l_dy)));

            s_slow_dx:= s_slow_dx+ l_dx;
            s_slow_dy:= s_slow_dy+ l_dy;

            if (Abs(s_slow_dx div k_slow)+ Abs(s_slow_dy) div k_slow)> 0
              then begin
                  // -- reduce the real volume movement
                  l_dx:= s_slow_dx div k_slow; l_dy:= s_slow_dy div k_slow;

                  compute_cosines;
                  l_move_norm:= f_2d_norm(f_2d_point(l_dx, l_dy));

                  m_c_volume_list_ref.draw(m_c_perspective_ref, clWhite);

                  if m_direction= e_x_direction
                    then begin
                        m_c_selected_volume.translate(l_move_norm* l_cos_x/ 10, 0, 0);
                        display('dir_X '+ f_display_double(l_move_norm* l_cos_x, 2, 2));
                      end
                    else
                      if m_direction= e_y_direction
                        then begin
                            m_c_selected_volume.translate(0, l_move_norm* l_cos_y/ 10, 0);
                            display('dir_Y');
                          end
                        else begin
                            m_c_selected_volume.translate(0, 0, - l_dy/ 10);
                            display('dir_Z');
                          end;
                  // -- reset the cumulated move
                  s_slow_dx:= 0; s_slow_dy:= 0;

                  m_c_perspective_ref.draw_unit_axes(clBlack);
                  m_c_volume_list_ref.draw(m_c_perspective_ref, clBlack);
                end;

              s_mouse_start_x:= p_x; s_mouse_start_y:= p_y;

          end; // cumulate_move

        begin // exact_direction
          display('> exact_dir '+ IntToStr(p_x)+ ' '+ IntToStr(s_mouse_start_x));

          // move_direct;
          cumulate_move;

          display('< exact_dir');
        end; // exact_direction

      begin // handle_mouse_move
        if (ssLeft in p_shift_state) and (m_c_selected_volume<> Nil)
          then begin
              if m_direction= e_unknown_direction
                then // approximate_direction
                else exact_direction;
            end;
      end; // handle_mouse_move

    Destructor c_scene.Destroy;
      begin
        Inherited;
      end; // Destroy

    begin // u_c_scene
    end. // u_c_scene

