// 002 u_c_volume_list
// 26 may 2005

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_c_volume_list;
  interface
    uses Classes, u_c_basic_object
        , u_c_3d_vertex_list
        , u_c_perspective
        , u_3d_point
        , u_2d_point
        , u_c_facet_list
        ;

    type c_volume_list= // "volume" list
                        Class(c_basic_object)
                           m_c_volume_list: tStringList;

                           Constructor create_volume_list(p_name: String);

                           function f_volume_count: Integer;
                           function f_c_volume(p_volume_index: Integer): c_facet_list;
                           function f_index_of(p_volume_name: String): Integer;
                           function f_c_find_by_volume(p_volume_name: String): c_facet_list;
                           procedure add_volume(p_volume_name: String; p_c_volume: c_facet_list);
                           function f_c_add_volume(p_volume_name: String): c_facet_list;
                           procedure display_volume_list;

                           procedure draw(p_c_perspective: c_perspective; p_pen_color: Integer);

                           procedure add_cube(p_cube_name: String);
                           procedure add_pyramid(p_pyramid_name: String);
                           procedure add_horizontal_plane(p_plane_name: String);
                           procedure add_vertical_x0z_plane(p_plane_name: String);
                           procedure add_vertical_y0z_plane(p_plane_name: String);

                           function f_c_find_selected_volume(p_2d_mouse: t_2d_point;
                               p_c_perspective: c_perspective): c_facet_list;

                           Destructor Destroy; Override;
                         end; // c_volume_list

  implementation
    uses SysUtils, Graphics, u_c_display
        , u_double
        , u_create_volumes        
        ;

    // -- c_volume_list

    Constructor c_volume_list.create_volume_list(p_name: String);
      begin
        Inherited create_basic_object(p_name);

        m_c_volume_list:= tStringList.Create;
      end; // create_volume_line

    function c_volume_list.f_volume_count: Integer;
      begin
        Result:= m_c_volume_list.Count;
      end; // f_volume_count

    function c_volume_list.f_c_volume(p_volume_index: Integer): c_facet_list;
      begin
        Result:= c_facet_list(m_c_volume_list.Objects[p_volume_index]);
      end; //  f_c_volume

    function c_volume_list.f_index_of(p_volume_name: String): Integer;
      begin
        Result:= m_c_volume_list.IndexOf(p_volume_name);
      end; // f_index_of

    function c_volume_list.f_c_find_by_volume(p_volume_name: String): c_facet_list;
      var l_index_of: Integer;
      begin
        l_index_of:= f_index_of(p_volume_name);
        if l_index_of< 0
          then Result:= Nil
          else Result:= c_facet_list(m_c_volume_list.Objects[l_index_of]);
      end; // f_c_find_by_name

    procedure c_volume_list.add_volume(p_volume_name: String; p_c_volume: c_facet_list);
      begin
        m_c_volume_list.AddObject(p_volume_name, p_c_volume);
      end; // add_volume

    function c_volume_list.f_c_add_volume(p_volume_name: String): c_facet_list;
      begin
        Result:= c_facet_list.create_facet_list(p_volume_name);
        m_c_volume_list.AddObject(p_volume_name, Result);
      end; // f_c_add_volume

    procedure c_volume_list.display_volume_list;
      var l_volume_index: Integer;
      begin
        display(m_name+ ' '+ IntToStr(f_volume_count));

        for l_volume_index:= 0 to f_volume_count- 1 do
          // display(f_c_volume(l_volume_index).f_display_volume);
      end; // display_volume_list

    procedure c_volume_list.draw(p_c_perspective: c_perspective; p_pen_color: Integer);
      var l_volume_index: Integer;
      begin
        for l_volume_index:= 0 to f_volume_count- 1 do
          f_c_volume(l_volume_index).draw(p_c_perspective, p_pen_color);
      end; // draw

    procedure c_volume_list.add_cube(p_cube_name: String);
      begin
        add_volume(p_cube_name, f_c_cube(p_cube_name, f_3d_point(0, 0, 0),
            f_3d_point(10, 0, 0), f_3d_point(0, 10, 0), f_3d_point(0, 0, 10)));
      end; // add_cube

    procedure c_volume_list.add_pyramid(p_pyramid_name: String);
      begin
        add_volume(p_pyramid_name, f_c_pyramid(p_pyramid_name, f_3d_point(0, 0, 0),
            f_3d_point(0, 10, 0), f_3d_point(10* Sqrt(3)/ 2, 5, 0), f_3d_point(10* Sqrt(3)/ 6, 5, 10* Sqrt(3)/ 2)));
      end; // add_pyramid

    procedure c_volume_list.add_horizontal_plane(p_plane_name: String);
      begin
        add_volume(p_plane_name, f_c_plane(p_plane_name, f_3d_point(0, 0, 10),
            f_3d_point(0, 10, 10), f_3d_point(10, 10, 10)));
      end; // add_horizontal_plane

    procedure c_volume_list.add_vertical_x0z_plane(p_plane_name: String);
      begin
        add_volume(p_plane_name, f_c_plane(p_plane_name, f_3d_point(0, 10, 0),
            f_3d_point(10, 10, 0), f_3d_point(10, 10, 10)));
      end; // add_vertical_x0z_plane

    procedure c_volume_list.add_vertical_y0z_plane(p_plane_name: String);
      begin
        add_volume(p_plane_name, f_c_plane(p_plane_name,
            f_3d_point(10, 10, 10), f_3d_point(10, 10, 0), f_3d_point(10, 0, 0)));
      end; // add_vertical_y0z_plane

    function c_volume_list.f_c_find_selected_volume(p_2d_mouse: t_2d_point;
        p_c_perspective: c_perspective): c_facet_list;
      var l_volume_index: Integer;
      begin
        Result:= Nil;

        for l_volume_index:= 0 to f_volume_count- 1 do
          with f_c_volume(l_volume_index) do
          begin
            if f_is_selected(p_2d_mouse, p_c_perspective)
              then begin
                  Result:= f_c_self;
                  // display('sel_vol='+ m_name);
                  Break;
                end;
          end;
      end; // f_c_find_selected_volume

    Destructor c_volume_list.Destroy;
      var l_volume_index: Integer;
      begin
        for l_volume_index:= 0 to f_volume_count- 1 do
          f_c_volume(l_volume_index).Free;
        m_c_volume_list.Free;

        Inherited;
      end; // Destroy

    begin // u_c_volume_list
    end. // u_c_volume_list
