// 002 u_create_volumes
// 26 may 2005

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_create_volumes;
  interface
    uses Classes, u_c_basic_object
        , u_c_3d_vertex_list
        , u_c_perspective
        , u_3d_point
        , u_c_facet_list
        ;

    function f_c_cube(p_name: String;
        p_3d_origin, p_3d_x, p_3d_y, p_3d_z: t_3d_point): c_facet_list;
    function f_c_pyramid(p_name: String;
        p_3d_A, p_3d_B, p_3d_C, p_3d_top: t_3d_point): c_facet_list;
    function f_c_plane(p_name: String;
        p_3d_A, p_3d_B, p_3d_C: t_3d_point): c_facet_list;

  implementation
    uses SysUtils, Graphics
        , u_types_constants, u_c_display
        , u_double
        , u_2d_point
        ;

    // -- global

    function f_c_cube(p_name: String;
        p_3d_origin, p_3d_x, p_3d_y, p_3d_z: t_3d_point): c_facet_list;
        // -- no use to add origin (0, 0, 0)
      var l_3d_A, l_3d_B, l_3d_C, l_3d_D, l_3d_E, l_3d_F, l_3d_G, l_3d_h: t_3d_point;
      begin
        l_3d_A:= p_3d_origin; l_3d_B:= f_add_3d_point(l_3d_A, p_3d_y);
        l_3d_C:= f_add_3d_point(l_3d_B, p_3d_z); l_3d_D:= f_add_3d_point(l_3d_A, p_3d_z);

        l_3d_F:= f_add_3d_point(l_3d_B, p_3d_x); l_3d_G:= f_add_3d_point(l_3d_C, p_3d_x);
        l_3d_H:= f_add_3d_point(l_3d_D, p_3d_x); l_3d_E:= f_add_3d_point(l_3d_A, p_3d_x);

        Result:= c_facet_list.create_facet_list(p_name);

        with Result do
        begin
          with f_c_add_facet('face1', - 1), m_c_3d_vertex_list do
          begin
            f_c_add_3d_vertex('A', l_3d_A);
            f_c_add_3d_vertex('B', l_3d_B);
            f_c_add_3d_vertex('C', l_3d_C);
            f_c_add_3d_vertex('D', l_3d_D);
          end; // with f_c_add_facet

          with f_c_add_facet('face2', 1), m_c_3d_vertex_list do
          begin
            f_c_add_3d_vertex('B', l_3d_B);
            f_c_add_3d_vertex('C', l_3d_C);
            f_c_add_3d_vertex('G', l_3d_G);
            f_c_add_3d_vertex('F', l_3d_F);
          end; // with f_c_add_facet

          // -- front
          with f_c_add_facet('face3',  - 1), m_c_3d_vertex_list do
          begin
            f_c_add_3d_vertex('G', l_3d_G);
            f_c_add_3d_vertex('F', l_3d_F);
            f_c_add_3d_vertex('E', l_3d_E);
            f_c_add_3d_vertex('H', l_3d_H);
          end; // with f_c_add_facet

          // -- top
          with f_c_add_facet('face4', 1), m_c_3d_vertex_list do
          begin
            f_c_add_3d_vertex('H', l_3d_H);
            f_c_add_3d_vertex('G', l_3d_G);
            f_c_add_3d_vertex('C', l_3d_C);
            f_c_add_3d_vertex('D', l_3d_D);
          end; // with f_c_add_facet

          compute_facet_normals;
        end; // with p_c_facet_list
      end; // f_c_cube

    function f_c_pyramid(p_name: String;
        p_3d_A, p_3d_B, p_3d_C, p_3d_top: t_3d_point): c_facet_list;
      begin
        Result:= c_facet_list.create_facet_list(p_name);

        with Result do
        begin
          with f_c_add_facet('back', -1), m_c_3d_vertex_list do
          begin
            f_c_add_3d_vertex('A', p_3d_A);
            f_c_add_3d_vertex('B', p_3d_B);
            f_c_add_3d_vertex('D', p_3d_top);
          end; // with f_c_add_facet

          with f_c_add_facet('east', -1), m_c_3d_vertex_list do
          begin
            f_c_add_3d_vertex('B', p_3d_B);
            f_c_add_3d_vertex('C', p_3d_C);
            f_c_add_3d_vertex('D', p_3d_top);
          end; // with f_c_add_facet

          with f_c_add_facet('west', 1), m_c_3d_vertex_list do
          begin
            f_c_add_3d_vertex('A', p_3d_A);
            f_c_add_3d_vertex('C', p_3d_C);
            f_c_add_3d_vertex('D', p_3d_top);
          end; // with f_c_add_facet
          compute_facet_normals;
        end; // with p_c_facet_list
      end; // f_c_pyramid

    function f_c_plane(p_name: String;
        p_3d_A, p_3d_B, p_3d_C: t_3d_point): c_facet_list;
      begin
        Result:= c_facet_list.create_facet_list(p_name);

        with Result do
        begin
          with f_c_add_facet('facet', -1), m_c_3d_vertex_list do
          begin
            f_c_add_3d_vertex('A', p_3d_A);
            f_c_add_3d_vertex('B', p_3d_B);
            f_c_add_3d_vertex('C', p_3d_C);
            f_c_add_3d_vertex('D', f_add_3d_point(p_3d_A, f_subtract_3d_point(p_3d_C, p_3d_B)));
          end; // with f_c_add_facet

          compute_facet_normals;
        end; // with p_c_facet_list
      end; // f_c_plane

    begin // u_c_facet_list
    end. // u_c_facet_list
